//////////////////////////////////////////////////////////////////////
//                            CuthillMcKee 
//////////////////////////////////////////////////////////////////////


//  Code for a version of the Cuthill-McKee algorithm to reduce the bandwidth
//  for the matrix based on a mesh generated by the code triangle
//  by Jonathan Richard Shewchuk (http://www.cs.cmu.edu/~quake/triangle.html)
//
//  Version 1.0                                                              
//  June 26, 2000                                                            
//
//  Copyright 2000                                                           
//  Andreas Stahel
//  Hochschule fuer Technik und Architektur
//  Quellgasse 21
//  CH-2501 Biel
//  Switzerland
//  Andreas.Stahel@hta-bi.bfh.ch
// 
//  This program may be freely redistributed under the condition that the    
//    copyright notices are not removed, and no compensation is received. 
//    Private, research, and institutional use is free. You may distribute 
//    modified versions of this code UNDER THE CONDITION THAT THIS CODE AND 
//    ANY MODIFICATIONS MADE TO IT IN THE SAME FILE REMAIN UNDER COPYRIGHT 
//    OF THE ORIGINAL AUTHOR, BOTH SOURCE AND OBJECT CODE ARE MADE FREELY 
//    AVAILABLE WITHOUT CHARGE, AND CLEAR NOTICE IS GIVEN OF THE 
//    MODIFICATIONS.  Distribution of this code as part of a commercial 
//    system is permissible ONLY BY DIRECT ARRANGEMENT WITH THE AUTHOR.  
//    (If you are not directly supplying this code to a    
//    customer, and you are instead telling them how they can obtain it for  
//    free, then you are not required to make any arrangement with me.)      



#include <iostream>            // defines ostream
#include <list>
#include <vector>
#include <string>
#include <algorithm>
#include <unistd.h>              // needs getopt()
#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h>

//using namespace std

#define max(A,B) ((A)>(B)?(A):(B)))

class TNode
{
private:
  int     Nedges;            // current number of not yet covered edges
  std::list <int> neighbor;        // list of all neighbors of the node
public:
  TNode ();                 // TNode();
 ~TNode ();                 // clean up dynamic memory

  void assign(TNode old);      // copy the node information
  int  edges();                // return the current number of edges
  void edges(int e);           // set the current number of edges
  void addEdge(int n);         // add an edge to node n
  void removeEdge(int n);      // remove edge to node n
  int  removeNextEdge();       // get the number of the next edge, 
                               // returns 0 if no more edges are available
  void getEdgeList(int* vec);  // get vector containing the node numbers
}; 
// end of class TNode definition

TNode::TNode(){
  Nedges=0;
  neighbor.clear();
}

TNode::~TNode(){
  Nedges=0;
  neighbor.clear();
}

void TNode::assign(TNode old){
  Nedges=old.edges();
  neighbor=old.neighbor;
}

int TNode::edges(){
  return Nedges;
}

void TNode::edges(int e){
  Nedges=e;
}

void TNode::addEdge(int n){
  std::list<int>::iterator p;
  p=find(neighbor.begin(),neighbor.end(),n);
  if(p==neighbor.end()){
    Nedges++;
    neighbor.push_front(n) ;
  }
}

void TNode::removeEdge(int n){
  Nedges--;
  neighbor.remove(n) ;
}

int TNode::removeNextEdge(){
  int res;
  if(Nedges>0){
    Nedges--;
    res=neighbor.front();
    neighbor.pop_front();
  }
  else res=0;
return(res);
}

void TNode::getEdgeList(int* vec){  // get vector containing the node numbers
  //  for(int i=0;i<Nedges;i++){vec[i]=neighbor(i);}
std::list<int>::iterator from;
int i=0; 
 for (from = neighbor.begin(); from != neighbor.end(); ++from)
   {vec[i]=*from; i++;}
}


void printhelp(){
std::cout<< "CuthillMcKee  -v -h -sn -mn -Sfile filename \n"
 <<" Version 0.1 , by Andreas.Stahel@hta-bi.bfh.ch\n"
 <<"  -sn to use node with number n as starting node\n"
 <<"  -s0 to use all nodes with minimal order as starting nodes (default)\n"
 <<"  -Sfile read list of starting nodes from file\n"
 <<"  -mn use at most n starting nodes\n"
 <<"  -g  generate a filename.sparse file to be read by octave to\n"
 <<"      display the sparsity structure\n"
 <<"      Use the Octave code in sparseplot.m to display the figure\n"
 <<"  -h  help, show this message\n"
 <<"  -v  verbose output\n"
    ;
  exit(0);
}

// the only global variable
TNode* node;


int findbandwidth(int NumberNodes, TNode* nodeBak, int*old2new);
void CuthillMcKee(int startnode,int NumberNodes, 
                  TNode* node, int*old2new, int*new2old);

int main(int argc, char* argv[]){

  char* infilebase=NULL;

  // using getopt() Swan p 241
#define OPTIONS "vhgs:S:m:"      // list of all command line options
  bool vOption=false, sOption=false, mOption=false,
       gOption=false, SOption=false;  // flags
  char* smArg ;               // string to -s -m options 
  char* SArg=NULL ;          // string to -S option 
  int startnode=0;
  int maxStartNodes=0;

  char c; //returned by getopt
  while((c=getopt(argc,argv,OPTIONS))!= -1)
    switch(c){
    case 's': {sOption=true;
      smArg=optarg;
      startnode=atol(smArg);}break;
    case 'S': SOption=true;
      SArg=optarg;break;
    case 'm': mOption=true;
      smArg=optarg;
      maxStartNodes=atol(smArg);break;
    case 'v': vOption=true;;break;
    case 'g': gOption=true;;break;
    case 'h': printhelp();break;
    case '?': printhelp();}
  if(argc!=optind+1){printhelp();}
  else{  infilebase=argv[optind];}

  std::string filename;
  filename=infilebase;
  filename.append(".node");
  std::ifstream infileNode(filename.c_str(),std::ios::in);
  if(!infileNode){std::cerr<<"Error: unable to open input file for nodes"
        <<std::endl;}
  
  int NumberNodes=0;
  infileNode>>NumberNodes;
  if(vOption){std::cout<<"Number of Nodes: "<<NumberNodes<<std::endl;}
  infileNode.close();
  if(!mOption) { maxStartNodes=NumberNodes; }

  int* old2new;
  int* old2newBest;
  int* new2old;
  int oldband=0;
  old2new=new int[NumberNodes+1];
  old2newBest=new int[NumberNodes+1];
  new2old=new int[NumberNodes+1];
  int bestband = NumberNodes;
  // generate the array to hold the structure

  //TNode* node; // defined as global variable now
  node = new TNode[NumberNodes+1];
  TNode* nodeBak;
  nodeBak = new TNode[NumberNodes+1];

  // read the information from the file into nodeBak
  filename=infilebase;
  filename.append(".ele");
  std::ifstream infileEle(filename.c_str(),std::ios::in);
  if(!infileEle){std::cerr<<"Error: unable to open input file for elements"
     <<std::endl;}

  int NumberElements=0;
  int NumberNodesEle=0;
  int NumberAttrib=0;
  infileEle>>NumberElements;  
  infileEle>>NumberNodesEle;
  infileEle>>NumberAttrib;
  int* tmpEle;
  tmpEle=new int[NumberNodesEle];

  for(int i=1;i<=NumberElements;i++){
    infileEle>>i;
    for(int j=0;j<NumberNodesEle;j++){infileEle>>tmpEle[j];}
    for(int ii=0;ii<NumberNodesEle;ii++){
      for(int jj=ii+1;jj<NumberNodesEle;jj++){
	nodeBak[tmpEle[ii]].addEdge(tmpEle[jj]);
	nodeBak[tmpEle[jj]].addEdge(tmpEle[ii]);
//	oldband=max(oldband,abs(tmpEle[ii]-tmpEle[jj]));
        int tmpmax=abs(tmpEle[ii]-tmpEle[jj]);
        if (tmpmax>oldband){oldband=tmpmax;}
      }
    }
  }
  infileEle.close();

  if(vOption){std::cout<<"Original semibandwidth: "<<oldband<<std::endl;}

  if(gOption){
    // count the number of connections to be stored
    TNode tempNode;
    int nnode;
    int count=0;
    for(int i=1;i<=NumberNodes;i++){
      tempNode.assign(nodeBak[i]);
      while( (nnode=tempNode.removeNextEdge())>0){
	if(i<nnode){ count++;}
      }
    }
    // write the result to the file
    filename=infilebase;
    filename.append(".sparse");
    std::ofstream outfile(filename.c_str(),std::ios::out);
    if(!outfile)
      {std::cerr<<"Error: unable to open output file for sparse structure"
          <<std::endl;}
    outfile<<NumberNodes<<"  "<<count<<std::endl;
    for(int i=1;i<=NumberNodes;i++){
      tempNode.assign(nodeBak[i]);
      while( (nnode=tempNode.removeNextEdge())>0){
	if(i<nnode){ outfile<< i <<" "<<nnode<<std::endl;}
      }
    }
    outfile.close();
  }//


  //choose the initial node(s) for Cuthill McKee

  std::list<int> StartingNodes;
  if(sOption&&(startnode!=0)) {StartingNodes.push_front(startnode);}
  // -s0 use all nodes of minimal order, then startnode was set to 0
  else{
    if(SOption){
      std::ifstream SNodefile(SArg,std::ios::in);
      if(!SNodefile)
	{std::cerr<<"Error: unable to open input file for starting nodes"
           <<std::endl;
	exit(1);}
      for(SNodefile>>startnode;!SNodefile.eof();SNodefile>>startnode) {
  	StartingNodes.push_back(startnode);
      }
      SNodefile.close();
    }
    else{
      // create a list of all nodes with minimal number of edges
      int minedge=10000;
      for(int i=1;i<=NumberNodes;i++){
	if (nodeBak[i].edges()==minedge) {StartingNodes.push_back(i) ;}
	else{
	  if (nodeBak[i].edges()<minedge) {
	    StartingNodes.clear();
	    minedge=nodeBak[i].edges();
	    StartingNodes.push_back(i) ; 
	  }
	}
      }
    }
  }
  if(!mOption){maxStartNodes=StartingNodes.size();}

  while((StartingNodes.size()>0)&&(maxStartNodes>0)){
    startnode=StartingNodes.front();
    if((startnode<1)||(startnode>NumberNodes)){
      std::cerr<<"illegal value for start node :"<<startnode
         <<std::endl<<std::endl;
      printhelp();
    }
    maxStartNodes--;
    if(vOption) std::cout<<"Starting node : "<<startnode<<"   ";
    
    // copy structure onto node from variable nodeBak
    for(int i=1;i<=NumberNodes;i++){ node[i].assign(nodeBak[i]);}
    for(int i=0;i<=NumberNodes;i++){old2new[i]=0;new2old[i]=0;}
    CuthillMcKee(startnode, NumberNodes, node, old2new, new2old);

    int newband=findbandwidth(NumberNodes, nodeBak, old2new);
    if(vOption){
      std::cout<<"semibandwidth: "<<newband<<std::endl;
    }
    if(newband<bestband){
      bestband=newband;
      for(int i=1;i<=NumberNodes;i++) old2newBest[i]=old2new[i];
    }
    StartingNodes.pop_front();
  }

// generate the map file
  if(vOption) std::cout<<"Final bandwidth: "<<bestband<<std::endl;
  filename=infilebase;
  filename.append(".map");
  std::ofstream outfileMap(filename.c_str(),std::ios::out);
  if(!outfileMap)
    {std::cerr<<"Error: unable to open output file for renumbering"
         <<std::endl;}
  outfileMap<<NumberNodes<<std::endl;
  for(int i=1;i<=NumberNodes;i++){
    outfileMap <<i<<" "<<old2newBest[i]<<std::endl;
  }
  outfileMap<<"# Generated by  ";
  for(int i=0;i<argc;i++){outfileMap<<argv[i]<<" ";}
  outfileMap<<std::endl;
  outfileMap.close();


//    for(int i=1;i<=NumberNodes;i++){
//      cout<<i<<" ";
//      node[i].~TNode();
//      //    cout<<": ";
//      //    nodeBak[i].~TNode();
//    }

  if (node != NULL){  delete [] node ; node=NULL;}
  if (nodeBak != NULL){  delete [] nodeBak ; node=NULL;}
  if (tmpEle != NULL){  delete [] tmpEle ; tmpEle=NULL;}
  if (old2new != NULL){  delete [] old2new ; old2new=NULL;}
  if (new2old != NULL){  delete [] new2old ; new2old=NULL;}
  if (old2newBest != NULL){  delete [] old2newBest ; old2newBest=NULL;}

  return 0;
}



int findbandwidth(int NumberNodes, TNode* nodeBak, int*old2new){
  int newband=0;
  int tmpband=0;
  int* nodelist=NULL;
  int size;
  for(int i=1;i<=NumberNodes;i++){
    size=nodeBak[i].edges();
    nodelist=new int[size];
    nodeBak[i].getEdgeList(nodelist);
    for(int j=0;j<size;j++){
//      newband=max(newband,abs(old2new[nodelist[j]]-old2new[i]));
        tmpband=abs(old2new[nodelist[j]]-old2new[i]);
        if (tmpband>newband){newband=tmpband;}
    }
    delete [] nodelist;
  }
  return(newband);
}

class compare{
    public: bool operator()(int a, int b) 
      { return node[a].edges() < node[b].edges(); }
//    {return a < b; }
};



void CuthillMcKee(int startnode,int NumberNodes,
                  TNode* node, int*old2new, int*new2old){
  int currnode=1;
  int worknode=1;
  int nnode;       // internal numbering of nodes connected to the current node
//m  list<int> NodeList;
  std::vector<int> NodeList;

  new2old[1]=startnode;
  old2new[startnode]=1;

  
  while(currnode<NumberNodes){  // run through all nodes
    // find all nodes attached to new2old[worknode]
    NodeList.clear();
    while( (nnode=node[new2old[worknode]].removeNextEdge())>0){
      node[nnode].removeEdge(new2old[worknode]);
//m      NodeList.push_front(nnode);
      NodeList.push_back(nnode);
    }
//m    NodeList.sort();  // wozu sortieren nach Nummerierung??
//    sort(NodeList.begin(),NodeList.end(),compare());
    // eventuell nach Anzahl freier Nachbarn sortieren ??
    // assign the new numbering
    while(NodeList.size()>0){
//m      nnode=NodeList.front();
      nnode=NodeList.back();
      if(old2new[nnode]==0){
	currnode++;
	new2old[currnode]=nnode;
	old2new[nnode]=currnode;
      }
//m      NodeList.pop_front();
      NodeList.pop_back();
    }//while NodeList
    worknode++;
  }//while currnode
}








